import unittest
import sys

if __name__ == '__main__':
    import xmlrunner
    loader = unittest.TestLoader()
    suite = loader.discover(start_dir='linstor_tests')
    runner = xmlrunner.XMLTestRunner(output='test-reports')
    res = runner.run(suite)
    sys.exit(1 if len(res.errors) > 0 else 0)



import unittest
from linstor import SizeCalc, LinstorError


class TestUtils(unittest.TestCase):
    def assertSizeUnit(self, val, exp_size, exp_unit):
        size, unit = SizeCalc.parse_unit(val)
        self.assertEqual(exp_size, size)
        self.assertEqual(exp_unit, unit)

    def test_size_calc_units(self):
        self.assertSizeUnit("128k", 128, SizeCalc.UNIT_KiB)
        self.assertSizeUnit("128", 128, SizeCalc.UNIT_B)
        self.assertSizeUnit("248GiB", 248, SizeCalc.UNIT_GiB)
        self.assertSizeUnit("248GB", 248, SizeCalc.UNIT_GB)

        self.assertRaises(LinstorError, SizeCalc.parse_unit, "nosize")

    def test_size_calc_convert(self):
        self.assertEqual(10485760, SizeCalc.auto_convert("10M", SizeCalc.UNIT_B))
        self.assertEqual(10485760, SizeCalc.auto_convert("10MiB", SizeCalc.UNIT_B))
        self.assertEqual(10000000, SizeCalc.auto_convert("10MB", SizeCalc.UNIT_B))

        self.assertEqual(3221225472, SizeCalc.auto_convert("3Gib", SizeCalc.UNIT_B))
        self.assertEqual(3145728, SizeCalc.auto_convert("3Gib", SizeCalc.UNIT_KiB))



"""
Linstor response module

Contains various classes of linstorapi responses wrappers.
"""
from datetime import datetime

import linstor.sharedconsts as apiconsts
from .errors import LinstorError


class RESTMessageResponse(object):
    """
    A base protobuf wrapper class, all api response use.
    """

    def __init__(self, rest_data):
        self._rest_data = rest_data

    def data(self, version):
        """
        Returns a specific version data format.
        :param str version:
        :return:
        """
        if version == "v0":
            return self.data_v0
        return self.data_v1

    @property
    def data_v0(self):
        return self._rest_data

    @property
    def data_v1(self):
        return self._rest_data

    @classmethod
    def _unix_epoch_field_value(cls, unix_epoch_val):
        dt = datetime.fromtimestamp(unix_epoch_val / 1000)
        return dt.replace(microsecond=(unix_epoch_val % 1000) * 1000)

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return True

    def __str__(self):
        return str(self._rest_data)

    def __repr__(self):
        return "RESTMessageResponse(" + repr(self._rest_data) + ")"


class ApiCallResponse(RESTMessageResponse):
    """
    This is a wrapper class for a proto MsgApiCallResponse.
    It provides some additional methods for easier state checking of the ApiCallResponse.
    """

    def __init__(self, rest_data):
        super(ApiCallResponse, self).__init__(rest_data)

    @classmethod
    def from_json(cls, json_data):
        """
        Creates a ApiCallResponse from a data block.

        :param json_data: Parsed json data with "ret_code", "message" and "details" fields.
        :return: a new ApiCallResponse()
        """
        return ApiCallResponse(json_data)

    @classmethod
    def from_str(cls, message, code=None):
        return ApiCallResponse({"message": message, "ret_code": apiconsts.FAIL_UNKNOWN_ERROR})

    def is_error(self, code=None):
        """
        Returns True if the ApiCallResponse is any error and "code" is unset.
        If "code" is set, return True if the given "code" matches the response code.

        :return: True if it is any error and "code" unset.
                 If "code" is set return True if "code" matches
                 response code. In any other cases (e.g., not an error at all), return False.
        """
        if self.ret_code & apiconsts.MASK_ERROR != apiconsts.MASK_ERROR:
            return False  # not an error at all

        return ((code & self.ret_code) == code) if code else True

    def is_warning(self):
        """
        Returns True if the ApiCallResponse is a warning.

        :return: True if it is a warning.
        """
        return True if self.ret_code & apiconsts.MASK_WARN == apiconsts.MASK_WARN else False

    def is_info(self):
        """
        Returns True if the ApiCallResponse is an info.

        :return: True if it is an info.
        """
        return True if self.ret_code & apiconsts.MASK_INFO == apiconsts.MASK_INFO else False

    def is_success(self):
        """
        Returns True if the ApiCallResponse is a success message.

        :return: True if it is a success message.
        """
        return not self.is_error() and not self.is_warning() and not self.is_info()

    @property
    def ret_code(self):
        """
        Returns the numeric return code mask.

        :return: Return code mask value
        """
        return self._rest_data["ret_code"]

    @property
    def message(self):
        return self._rest_data.get("message")

    @property
    def cause(self):
        return self._rest_data.get("cause")

    @property
    def correction(self):
        return self._rest_data.get("correction")

    @property
    def details(self):
        return self._rest_data.get("details")

    @property
    def object_refs(self):
        """
        Returns a dict generator with the object_references.

        :return: Dict with object references
        :rtype: dict[str, str]
        """
        return self._rest_data.get("obj_refs", {})

    @property
    def error_report_ids(self):
        return self._rest_data.get("error_report_ids", [])

    @property
    def data_v0(self):
        d = {
            "ret_code": self.ret_code,
            "message": self.message
        }
        if self.cause:
            d["cause"] = self.cause
        if self.correction:
            d["correction"] = self.correction
        if self.details:
            d["details"] = self.details
        if self.object_refs:
            d["object_refs"] = [{"key": x, "value": self.object_refs[x]} for x in self.object_refs]
        if self.error_report_ids:
            d["error_report_ids"] = self.error_report_ids
        return d

    def __eq__(self, other):
        if isinstance(other, ApiCallResponse):
            return self.__hash__() == other.__hash__()
        return False

    def __hash__(self):
        return hash((self.ret_code, self.message))

    def __str__(self):
        st_str = "SUCC"
        if self.is_error():
            st_str = "ERRO"
        elif self.is_info():
            st_str = "INFO"
        elif self.is_warning():
            st_str = "WARN"

        return st_str + ":" + self.message


class ErrorReport(RESTMessageResponse):
    def __init__(self, data):
        super(ErrorReport, self).__init__(data)

    @property
    def datetime(self):
        dt = datetime.fromtimestamp(self._rest_data["error_time"] / 1000)
        return dt.replace(microsecond=(self._rest_data["error_time"] % 1000) * 1000)

    @property
    def id(self):
        return self._rest_data['filename'][len("ErrorReport-"):-len(".log")]

    @property
    def text(self):
        return self._rest_data.get("text")

    @property
    def node_name(self):
        return self._rest_data.get("node_name")

    @property
    def node_names(self):
        return self._rest_data.get("node_name")

    @property
    def module(self):
        return self._rest_data.get("module", "")

    @property
    def version(self):
        return self._rest_data.get("version", "")

    @property
    def peer(self):
        return self._rest_data.get("peer", "")

    @property
    def exception(self):
        return self._rest_data.get("exception", "")

    @property
    def exception_message(self):
        return self._rest_data.get("exception_message", "")

    @property
    def origin_file(self):
        return self._rest_data.get("origin_file", "")

    @property
    def origin_line(self):
        """

        :return: origin line of the exception
        :rtype: Optional[int]
        """
        return int(self._rest_data["origin_line"]) if "origin_line" in self._rest_data else None

    @property
    def data_v0(self):
        d = self._rest_data
        d["node_names"] = d["node_name"]
        del d["node_name"]
        return d


class NodeType(object):
    CONTROLLER = apiconsts.VAL_NODE_TYPE_CTRL
    SATELLITE = apiconsts.VAL_NODE_TYPE_STLT
    COMBINED = apiconsts.VAL_NODE_TYPE_CMBD
    AUXILIARY = apiconsts.VAL_NODE_TYPE_AUX


class NetInterface(RESTMessageResponse):
    def __init__(self, rest_data):
        super(NetInterface, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data["name"]

    @property
    def address(self):
        return self._rest_data["address"]

    @property
    def stlt_port(self):
        return self._rest_data.get("satellite_port")

    @property
    def stlt_encryption_type(self):
        return self._rest_data.get("satellite_encryption_type")

    @property
    def is_active(self):
        return self._rest_data.get("is_active")

    @property
    def data_v0(self):
        return {
            "address": self.address,
            "name": self.name,
            "stlt_port": self.stlt_port,
            "stlt_encryption_type": self.stlt_encryption_type
        }


class Node(RESTMessageResponse):
    def __init__(self, rest_data):
        super(Node, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data["name"]

    @property
    def type(self):
        return self._rest_data["type"]

    @property
    def connection_status(self):
        return self._rest_data.get(
            "connection_status",
            apiconsts.ConnectionStatus.OFFLINE.name)

    @property
    def net_interfaces(self):
        return [NetInterface(x) for x in self._rest_data.get("net_interfaces", [])]

    @property
    def storage_providers(self):
        return self._rest_data.get("storage_providers", [])

    @property
    def resource_layers(self):
        return self._rest_data.get("resource_layers", [])

    @property
    def unsupported_providers(self):
        return self._rest_data.get("unsupported_providers", {})

    @property
    def unsupported_layers(self):
        return self._rest_data.get("unsupported_layers", {})

    @property
    def props(self):
        return self._rest_data.get("props", {})

    @property
    def properties(self):
        return self._rest_data.get("props", {})

    @property
    def flags(self):
        return self._rest_data.get("flags", [])

    @property
    def eviction_timestamp(self):
        return self._rest_data.get("eviction_timestamp")

    @property
    def data_v0(self):
        d = dict(self._rest_data)
        d["props"] = [{"key": x, "value": self.props[x]} for x in self.props]
        d["connection_status"] = apiconsts.ConnectionStatus[
            self._rest_data.get("connection_status", apiconsts.ConnectionStatus.UNKNOWN.name)].value
        d["net_interfaces"] = [x.data_v0 for x in self.net_interfaces]
        return d

    def __str__(self):
        return "Node({n}, {t}, {con})".format(n=self.name, t=self.type, con=self.connection_status)


class NodeListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(NodeListResponse, self).__init__(rest_data)

    @property
    def nodes(self):
        """
        Returns a list with all nodes.

        :return: The node list.
        :rtype: list[Node]
        """
        return [Node(x) for x in self._rest_data]

    def node(self, node_name):
        """
        Returns the specified node from the nodelist.

        :param str node_name: Node name
        :return: Node object of the node, or None
        :rtype: Node
        """
        for n in self.nodes:
            if n.name == node_name:
                return n
        return None

    @property
    def data_v0(self):
        return {
            "nodes": [n.data_v0 for n in self.nodes]
        }


class FreeSpace(RESTMessageResponse):
    def __init__(self, rest_data):
        super(FreeSpace, self).__init__(rest_data)

    @property
    def free_capacity(self):
        return self._rest_data.get("free_capacity")

    @property
    def total_capacity(self):
        return self._rest_data.get("total_capacity")

    def __str__(self):
        return "{used}/{total} Kib used".format(
            used=self.total_capacity - self.free_capacity, total=self.total_capacity)

    @property
    def data_v0(self):
        return {
            "stor_pool_name": self._rest_data["storage_pool_name"],
            "free_capacity": self.free_capacity,
            "total_capacity": self.total_capacity
        }


class StoragePoolDriver(object):
    LVM = "LVM"
    LVMThin = "LVM_THIN"
    ZFS = "ZFS"
    ZFSThin = "ZFS_THIN"
    Diskless = "DISKLESS"
    FILE = "FILE"
    FILEThin = "FILE_THIN"
    SPDK = "SPDK"
    REMOTE_SPDK = "REMOTE_SPDK"
    OPENFLEX_TARGET = "OPENFLEX_TARGET"
    EXOS = "EXOS"
    STORAGE_SPACES = "STORAGE_SPACES"
    STORAGE_SPACES_THIN = "STORAGE_SPACES_THIN"
    EBS_TARGET = "EBS_TARGET"
    EBS_INIT = "EBS_INIT"

    @staticmethod
    def list():
        return [
            StoragePoolDriver.LVM,
            StoragePoolDriver.LVMThin,
            StoragePoolDriver.ZFS,
            StoragePoolDriver.ZFSThin,
            StoragePoolDriver.Diskless,
            StoragePoolDriver.FILE,
            StoragePoolDriver.FILEThin,
            StoragePoolDriver.SPDK,
            StoragePoolDriver.REMOTE_SPDK,
            StoragePoolDriver.OPENFLEX_TARGET,
            StoragePoolDriver.EXOS,
            StoragePoolDriver.STORAGE_SPACES,
            StoragePoolDriver.STORAGE_SPACES_THIN,
            StoragePoolDriver.EBS_TARGET,
            StoragePoolDriver.EBS_INIT
        ]

    @classmethod
    def diskless_driver(cls):
        return [
            StoragePoolDriver.Diskless,
            StoragePoolDriver.EBS_INIT
        ]

    @staticmethod
    def storage_driver_pool_to_props(storage_driver, driver_pool_name):
        if storage_driver in [StoragePoolDriver.Diskless]:
            return {}

        if not driver_pool_name:
            raise LinstorError(
                "Driver '{drv}' needs a driver pool name.".format(drv=storage_driver)
            )

        if storage_driver == StoragePoolDriver.LVM:
            return {apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_VOLUME_GROUP: driver_pool_name}

        if storage_driver == StoragePoolDriver.LVMThin:
            driver_pool_parts = driver_pool_name.split('/')
            if not len(driver_pool_parts) == 2:
                raise LinstorError("Pool name '{dp}' does not have format VG/LV".format(dp=driver_pool_name))
            return {
                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_VOLUME_GROUP: driver_pool_parts[0],
                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_THIN_POOL: driver_pool_parts[1]
            }

        if storage_driver == StoragePoolDriver.ZFS:
            return {apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_ZPOOL: driver_pool_name}

        if storage_driver == StoragePoolDriver.ZFSThin:
            return {apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_ZPOOLTHIN: driver_pool_name}

        if storage_driver in [
                StoragePoolDriver.FILE,
                StoragePoolDriver.FILEThin]:
            return {apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_FILE_DIRECTORY: driver_pool_name}

        if storage_driver == StoragePoolDriver.SPDK:
            return {apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_VOLUME_GROUP: driver_pool_name}

        if storage_driver == StoragePoolDriver.OPENFLEX_TARGET:
            return {
                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_OPENFLEX_STOR_POOL: driver_pool_name
            }

        if storage_driver == StoragePoolDriver.STORAGE_SPACES:
            return {
                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_STORAGE_SPACES: driver_pool_name
            }

        if storage_driver == StoragePoolDriver.STORAGE_SPACES_THIN:
            return {
                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_STORAGE_SPACES: driver_pool_name
            }

        raise LinstorError(
            "Unknown storage driver '{drv}', known drivers: "
            "lvm, lvmthin, zfs, diskless, spdk".format(drv=storage_driver)
        )

    @classmethod
    def storage_props_to_driver_pool(cls, storage_driver, props):
        """
        Find the storage pool value for the given storage_driver in the given props.

        :param str storage_driver: String specifying a storage driver [``Lvm``, ``LvmThin``, ``Zfs``]
        :param props: Properties to search the storage pool value.
        :return: If found the storage pool value, else ''
        :rtype: str
        """
        if apiconsts.NAMESPC_STORAGE_DRIVER + '/StorPoolName' in props:
            return props[apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_NAME]

        storage_driver_enum = storage_driver
        if storage_driver_enum == StoragePoolDriver.LVM:
            return props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_VOLUME_GROUP, '')

        if storage_driver_enum == StoragePoolDriver.LVMThin:
            vg = props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_VOLUME_GROUP, '')
            lv = props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_THIN_POOL, '')
            return "{vg}/{lv}".format(vg=vg, lv=lv)

        if storage_driver_enum == StoragePoolDriver.ZFS:
            return props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_ZPOOL, '')

        if storage_driver_enum == StoragePoolDriver.ZFSThin:
            return props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_ZPOOLTHIN, '')

        if storage_driver_enum == StoragePoolDriver.SPDK:
            return props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_VOLUME_GROUP, '')

        if storage_driver_enum == StoragePoolDriver.OPENFLEX_TARGET:
            return props.get(apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_OPENFLEX_STOR_POOL, '')

        return ''


class StoragePool(RESTMessageResponse):
    DRIVER_KIND_MAP = {
        "DISKLESS": "DisklessDriver",
        "LVM": "LvmDriver",
        "LVM_THIN": "LvmThinDriver",
        "ZFS": "ZfsDriver",
        "ZFS_THIN": "ZfsThinDriver",
        "SPDK": "SpdkDriver"
    }

    def __init__(self, rest_data):
        super(StoragePool, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data["storage_pool_name"]

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def node_name(self):
        """
        Node name where the storage pool is used
        :return: node name
        :rtype: str
        """
        return self._rest_data["node_name"]

    @property
    def driver(self):
        """
        Provider kind string
        :return: provider kind string
        :rtype: str
        """
        return self.provider_kind

    @property
    def provider_kind(self):
        """
        Provider kind string
        :return: provider kind string
        :rtype: str
        """
        return self._rest_data.get("provider_kind")

    @property
    def properties(self):
        """
        Storage pool properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def static_traits(self):
        """
        Static traits.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("static_traits", {})

    @property
    def free_space(self):
        """
        Returns the free space object of the storage pool
        :return:
        :rtype: FreeSpace
        """
        if "free_capacity" in self._rest_data:
            return FreeSpace(self._rest_data)
        return None

    @property
    def free_space_mgr_name(self):
        return self._rest_data.get("free_space_mgr_name")

    def supports_snapshots(self):
        sup_snaps = self._rest_data.get("supports_snapshots")
        return self.static_traits.get("SupportsSnapshots", "false") == "true" if sup_snaps is None else sup_snaps

    def is_thin(self):
        """
        Checks if pool is thin
        :return: True if it is a thin pool
        :rtype: bool
        """
        return self.static_traits.get("Provisioning", "") == "Thin"

    def is_fat(self):
        """
        Checks if pool is fat
        :return: True if it is a fat pool
        :rtype: bool
        """
        return self.static_traits.get("Provisioning", "") == "Fat"

    def is_diskless(self):
        """
        Checks if pool is diskless
        :return: True if it is a diskless pool
        :rtype: bool
        """
        return self.provider_kind in StoragePoolDriver.diskless_driver()

    @property
    def reports(self):
        return [ApiCallResponse(x) for x in self._rest_data.get("reports", [])]

    @property
    def data_v0(self):
        d = {
            "stor_pool_uuid": self.uuid,
            "stor_pool_name": self.name,
            "node_name": self.node_name,
            "free_space_mgr_name": self.free_space_mgr_name
        }
        if self.free_space:
            d["free_space"] = self.free_space.data_v0
        d['driver'] = self.DRIVER_KIND_MAP.get(self.provider_kind, '')
        d['static_traits'] = [{"key": x, "value": v} for x, v in self.static_traits.items()]
        if self.properties:
            d["props"] = [{"key": x, "value": v} for x, v in self.properties.items()]
        return d


class StoragePoolListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(StoragePoolListResponse, self).__init__(rest_data)

    @property
    def storage_pools(self):
        """
        Returns list of storage pool objects.
        :return: list of storage pools
        :rtype: list[StoragePool]
        """
        return [StoragePool(x) for x in self._rest_data]

    @property
    def data_v0(self):
        return {
            "stor_pools": [x.data_v0 for x in self.storage_pools]
        }


class KeyValueStoresResponse(RESTMessageResponse):
    def __init__(self, data):
        super(KeyValueStoresResponse, self).__init__(data)

    def instances(self):
        """
        Returns a list of all known instances
        :return: List with all names of instances
        :rtype: list[str]
        """
        return [x['name'] for x in self._rest_data]

    def instance(self, name):
        """
        Returns a KeyValueStore object containing the specified KV instance.

        :param str name: name of the instance wanted
        :return: KeyValueStore object of the instance, if none found an empty is created
        :rtype: KeyValueStore
        """
        kv = [x for x in self._rest_data if x['name'] == name]
        kv = kv[0] if kv else {}
        return KeyValueStore(name, kv.get('props', {}))


class KeyValueStore(object):
    def __init__(self, instance_name, props):
        self._instance_name = instance_name
        self._props = props

    @property
    def properties(self):
        """
        Returns the property dictionary.

        :return: dict containing key values
        :rtype: dict[str, str]
        """
        return self._props

    def __str__(self):
        return str({"name": self._instance_name, "properties": self._props})


class DrbdVolumeDefinitionData(RESTMessageResponse):
    def __init__(self, rest_data):
        super(DrbdVolumeDefinitionData, self).__init__(rest_data)

    @property
    def resource_name_suffix(self):
        return self._rest_data["rsc_name_suffix"]

    @property
    def minor(self):
        return self._rest_data["minor_number"]

    @property
    def number(self):
        return self._rest_data["volume_number"]


class VolumeDefinition(RESTMessageResponse):
    def __init__(self, rest_data):
        super(VolumeDefinition, self).__init__(rest_data)

    @property
    def number(self):
        """
        Volume definition number

        :return: Volume definition number
        :rtype: int
        """
        return self._rest_data["volume_number"]

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def size(self):
        """
        Nett volume size in KiB.

        :return: Nett volume size in KiB.
        :rtype: int
        """
        return self._rest_data["size_kib"]

    @property
    def flags(self):
        """
        Resource definition flags as string list.

        :return: Resource definition flags as string list
        :rtype: list[str]
        """
        return self._rest_data.get("flags", [])

    @property
    def properties(self):
        """
        Resource definition properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def drbd_data(self):
        for layer in self._rest_data.get("layer_data", []):
            if layer["type"] == "DRBD" and layer.get("data"):
                return DrbdVolumeDefinitionData(layer["data"])
        return None

    @property
    def data_v0(self):
        """
        Returns compatibility output for the first machine readable format.

        :return: Dictionary with old resource definition format
        """
        v0_vlm_dfn = {
            "vlm_nr": self.number,
            "vlm_size": self.size,
            "vlm_dfn_uuid": self.uuid
        }

        if self.flags:
            v0_vlm_dfn['vlm_flags'] = self.flags

        if self.properties:
            v0_vlm_dfn['vlm_props'] = [{"key": x, "value": v} for x, v in self.properties.items()]

        drbd_data = self.drbd_data
        if drbd_data:
            v0_vlm_dfn['vlm_minor'] = drbd_data.minor

        return v0_vlm_dfn


class VolumeDefinitionResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(VolumeDefinitionResponse, self).__init__(rest_data)

    @property
    def volume_definitions(self):
        """

        :return:
        :rtype: list[VolumeDefinition]
        """
        return [VolumeDefinition(x) for x in self._rest_data]

    @property
    def rest_data(self):
        return self._rest_data


class DrbdLayer(RESTMessageResponse):
    def __init__(self, rest_data):
        super(DrbdLayer, self).__init__(rest_data)

    @property
    def port(self):
        return self._rest_data["port"]

    @property
    def secret(self):
        return self._rest_data["secret"]


class ResourceDefinition(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ResourceDefinition, self).__init__(rest_data)

    @property
    def name(self):
        """
        Resource definition name.

        :return: Resource definition name
        :rtype: str
        """
        return self._rest_data["name"]

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def external_name(self):
        """
        Returns the external name of the resource
        :return:
        :rtype: str
        """
        return self._rest_data.get("external_name", "")

    @property
    def flags(self):
        """
        Resource definition flags as string list.

        :return: Resource definition flags as string list
        :rtype: list[str]
        """
        return self._rest_data.get("flags", [])

    @property
    def properties(self):
        """
        Resource definition properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def drbd_data(self):
        for layer in self._rest_data.get("layer_data", []):
            if layer["type"] == "DRBD" and layer.get("data"):
                return DrbdLayer(layer["data"])
        return None

    @property
    def volume_definitions(self):
        """
        List of all volume definitions

        :return:
        :rtype: list[VolumeDefinition]
        """
        return [VolumeDefinition(x) for x in self._rest_data.get("volume_definitions", [])]

    @property
    def resource_group_name(self):
        """
        Returns the resource group name linked to the resource.

        :return: Name of the resource group this resource belongs too.
        :rtype: str
        """
        return self._rest_data.get("resource_group_name", "")


class ResourceDefinitionResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ResourceDefinitionResponse, self).__init__(rest_data)

    @property
    def resource_definitions(self):
        """
        List of resource definitions
        :return: List of resource definitions
        :rtype: list[ResourceDefinition]
        """
        return [ResourceDefinition(x) for x in self._rest_data]

    @property
    def data_v0(self):
        """
        Returns compatibility output for the first machine readable format.

        :return: Dictionary with old resource definition format
        """
        rsc_dfns = []
        for rsc_dfn in self.resource_definitions:
            v0_rsc_dfn = {
                "rsc_name": rsc_dfn.name,
                "rsc_dfn_uuid": rsc_dfn.uuid,
                "vlm_dfns": [x.data_v0 for x in rsc_dfn.volume_definitions]
            }

            if rsc_dfn.flags:
                v0_rsc_dfn['rsc_dfn_flags'] = rsc_dfn.flags
            if rsc_dfn.properties:
                v0_rsc_dfn["rsc_dfn_props"] = [{"key": x, "value": v} for x, v in rsc_dfn.properties.items()]

            drbd_data = rsc_dfn.drbd_data
            if drbd_data:
                v0_rsc_dfn['rsc_dfn_port'] = drbd_data.port
                v0_rsc_dfn['rsc_dfn_secret'] = drbd_data.secret

            rsc_dfns.append(v0_rsc_dfn)
        return {
            "rsc_dfns": rsc_dfns
        }


class SelectFilter(RESTMessageResponse):
    def __init__(self, rest_data):
        super(SelectFilter, self).__init__(rest_data)

    @property
    def place_count(self):
        return self._rest_data.get("place_count")

    @property
    def diskless_on_remaining(self):
        return self._rest_data.get("diskless_on_remaining", False)

    @property
    def storage_pool(self):
        return self._rest_data.get("storage_pool")

    @property
    def storage_pool_list(self):
        """
        Returns the list of storage pools used
        :return: storage pool list
        :rtype: List[str]
        """
        return self._rest_data.get("storage_pool_list", [self.storage_pool] if self.storage_pool else [])

    @property
    def not_place_with_rsc(self):
        return self._rest_data.get("not_place_with_rsc")

    @property
    def not_place_with_rsc_regex(self):
        return self._rest_data.get("not_place_with_rsc_regex")

    @property
    def replicas_on_same(self):
        return self._rest_data.get("replicas_on_same")

    @property
    def replicas_on_different(self):
        return self._rest_data.get("replicas_on_different")

    @property
    def layer_stack(self):
        return self._rest_data.get("layer_stack")

    @property
    def provider_list(self):
        return self._rest_data.get("provider_list")

    def __str__(self):
        fields = []
        if self.place_count:
            fields.append("PlaceCount: " + str(self.place_count))

        if self.storage_pool or self.storage_pool_list:
            fields.append("StoragePool(s): " + ", ".join(self.storage_pool_list))

        if "diskless_on_remaining" in self._rest_data:
            fields.append("DisklessOnRemaining: " + str(self.diskless_on_remaining))

        if self.not_place_with_rsc:
            fields.append("NotPlaceWithRsc: " + str(self.not_place_with_rsc))

        if self.not_place_with_rsc_regex:
            fields.append("NotPlaceWithRscRegex: " + str(self.not_place_with_rsc_regex))

        if self.replicas_on_same:
            fields.append("ReplicasOnSame: " + str(self.replicas_on_same))

        if self.replicas_on_different:
            fields.append("ReplicasOnDifferent: " + str(self.replicas_on_different))

        if self.layer_stack:
            fields.append("LayerStack: " + str(self.layer_stack))

        if self.provider_list:
            fields.append("ProviderList: " + str(self.provider_list))
        return "\n".join(fields)


class ResourceGroup(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ResourceGroup, self).__init__(rest_data)

    @property
    def name(self):
        """
        Resource group name.

        :return: Resource group name
        :rtype: str
        """
        return self._rest_data["name"]

    @property
    def description(self):
        """
        Resource group description.

        :return: Group description
        :rtype: str
        """
        return self._rest_data.get("description", "")

    @property
    def properties(self):
        """
        Resource group properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def select_filter(self):
        """
        Returns the select filter for the resource group.

        :return: Select filter class
        :rtype: SelectFilter
        """
        return SelectFilter(self._rest_data.get("select_filter", {}))


class ResourceGroupResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ResourceGroupResponse, self).__init__(rest_data)

    @property
    def resource_groups(self):
        """
        List of resource groups
        :return: List of resource groups
        :rtype: list[ResourceGroup]
        """
        return [ResourceGroup(x) for x in self._rest_data]


class VolumeGroup(RESTMessageResponse):
    def __init__(self, rest_data):
        super(VolumeGroup, self).__init__(rest_data)

    @property
    def number(self):
        """
        Volume number

        :return: volume number
        :rtype: int
        """
        return self._rest_data["volume_number"]

    @property
    def properties(self):
        """
        Volume group properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def flags(self):
        """
        Volume group flags.

        :return: Flags list
        :rtype:  list[str]
        """
        return self._rest_data.get("flags", [])


class VolumeGroupResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(VolumeGroupResponse, self).__init__(rest_data)

    @property
    def volume_groups(self):
        """
        List of volume groups
        :return: List of volume groups
        :rtype: list[VolumeGroup]
        """
        return [VolumeGroup(x) for x in self._rest_data]


class VolumeState(RESTMessageResponse):
    def __init__(self, data):
        super(VolumeState, self).__init__(data)

    @property
    def number(self):
        """
        Volume number index
        :return: Volume number index
        :rtype: int
        """
        return self._rest_data["vlm_nr"]

    @property
    def disk_state(self):
        """
        :return: String describing the disk state
        :rtype: str
        """
        return self._rest_data.get("disk_state")

    @property
    def data_v0(self):
        return {
            "vlm_nr": self.number,
            "disk_state": self.disk_state
        }


class ResourceState(RESTMessageResponse):
    def __init__(self, data):
        super(ResourceState, self).__init__(data)

    @property
    def name(self):
        return self._rest_data["rsc_name"]

    @property
    def rsc_name(self):
        return self.name

    @property
    def node_name(self):
        return self._rest_data["node_name"]

    @property
    def in_use(self):
        """
        Indicates if a resource is in use, for a drbd resource this means primary.
        Other types might be unknown/None
        :return: bool or None
        """
        return self._rest_data.get("in_use")

    @property
    def volume_states(self):
        """
        Returns volume states
        :return: volume states list
        :rtype: list[VolumeState]
        """
        return [VolumeState(x) for x in self._rest_data.get("vlm_states", [])]

    @property
    def data_v0(self):
        d = {
            "rsc_name": self.name,
            "node_name": self.node_name
        }

        if self.in_use is not None:
            d["in_use"] = self.in_use

        if self.volume_states:
            d["vlm_states"] = [x.data_v0 for x in self.volume_states]
        return d


class DrbdConnection(RESTMessageResponse):
    def __init__(self, rest_data):
        super(DrbdConnection, self).__init__(rest_data)

    @property
    def connected(self):
        return self._rest_data["connected"]

    @property
    def message(self):
        return self._rest_data.get("message")


class DrbdResource(RESTMessageResponse):
    def __init__(self, data):
        super(DrbdResource, self).__init__(data)

    @property
    def node_id(self):
        """
        Get DRBD node id
        :return: node id
        :rtype: int
        """
        return self._rest_data.get("node_id")

    @property
    def peer_slots(self):
        """
        Get DRBD peer slots
        :return: peer slot count
        :rtype: int
        """
        return self._rest_data.get("peer_slots")

    @property
    def al_stripes(self):
        """
        Get DRBD activity log stripes
        :return: al_stripes
        :rtype: int
        """
        return self._rest_data.get("al_stripes")

    @property
    def al_size(self):
        """
        Get DRBD activity log size
        :return: al size
        :rtype: int
        """
        return self._rest_data.get("al_size")

    @property
    def connections(self):
        """
        Connections dict of this DRBD Resource.

        :return: A node to DrbdConnection dict
        :rtype: dict[str, DrbdConnection]
        """
        return {k: DrbdConnection(v) for k, v in self._rest_data.get("connections", {}).items()}

    # TODO other fields


class ResourceLayerData(RESTMessageResponse):
    def __init__(self, data):
        super(ResourceLayerData, self).__init__(data)

    @property
    def name_suffix(self):
        return self._rest_data["rsc_name_suffix"]

    @property
    def children(self):
        """
        Return resource layer list children.
        :return: List of resource layer data children
        :rtype: list[ResourceLayerData]
        """
        return [ResourceLayerData(x) for x in self._rest_data.children]

    @property
    def type(self):
        return self._rest_data["type"]

    @property
    def drbd_resource(self):
        """
        Gets the DRBD resource layer data if layer data is DRBD, otherwise None.

        :return: None if it isn't a drbd resource, otherwise the DrbdResource object
        :rtype: Optional[DrbdResource]
        """
        if self.type == "DRBD":
            return DrbdResource(self._rest_data["drbd"])
        return None

    # TODO other layer objects


class VolumeLayerData(RESTMessageResponse):
    def __init__(self, data):
        super(VolumeLayerData, self).__init__(data)

    @property
    def layer_type(self):
        """
        Returns the name of the layer type.
        :return: Name of the layer type
        :rtype: str
        """
        return self._rest_data['type']


class DrbdVolumeDefinition(RESTMessageResponse):
    def __init__(self, data):
        super(DrbdVolumeDefinition, self).__init__(data)

    @property
    def number(self):
        return self._rest_data["volume_number"]

    @property
    def minor(self):
        return self._rest_data["minor_number"]

    @property
    def resource_name_suffix(self):
        return self._rest_data["resource_name_suffix"]


class DrbdVolumeData(RESTMessageResponse):
    def __init__(self, data):
        super(DrbdVolumeData, self).__init__(data)

    @property
    def drbd_volume_definition(self):
        return DrbdVolumeDefinition(self._rest_data["drbd_volume_definition"])

    @property
    def device_path(self):
        return self._rest_data["device_path"]

    @property
    def backing_device(self):
        return self._rest_data.get("backing_device")

    @property
    def meta_disk(self):
        return self._rest_data.get('meta_disk', "")

    @property
    def allocated_size(self):
        return self._rest_data["allocated_size_kib"]

    @property
    def usable_size(self):
        return self._rest_data["usable_size_kib"]


class StorageVolumeData(RESTMessageResponse):
    def __init__(self, data):
        super(StorageVolumeData, self).__init__(data)


class LUKSVolumeData(RESTMessageResponse):
    def __init__(self, data):
        super(LUKSVolumeData, self).__init__(data)


class Volume(RESTMessageResponse):
    def __init__(self, data):
        super(Volume, self).__init__(data)

    @property
    def number(self):
        return self._rest_data["volume_number"]

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def storage_pool_name(self):
        return self._rest_data.get("storage_pool_name")

    @property
    def storage_pool_driver_name(self):
        return self._rest_data["provider_kind"]

    @property
    def device_path(self):
        return self._rest_data.get("device_path")

    @property
    def allocated_size(self):
        return self._rest_data.get('allocated_size_kib')

    @property
    def usable_size(self):
        if self.drbd_data:
            return self.drbd_data.usable_size
        return None

    @property
    def flags(self):
        """
        Volume flags as string list.

        :return: Resource definition flags as string list
        :rtype: list[str]
        """
        return self._rest_data.get("flags", [])

    @property
    def properties(self):
        """
        Volume properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def layer_data(self):
        return [VolumeLayerData(x) for x in self._rest_data.get("layer_data", [])]

    @property
    def drbd_data(self):
        for layer in self._rest_data.get("layer_data_list", []):
            if layer["type"] == "DRBD" and layer.get("data"):
                return DrbdVolumeData(layer["data"])
        return None

    @property
    def storage_data(self):
        for layer in self._rest_data.get("layer_data_list", []):
            if layer["type"] == "STORAGE" and layer.get("data"):
                return DrbdVolumeData(layer["data"])
        return None

    @property
    def luks_data(self):
        for layer in self._rest_data.get("layer_data_list", []):
            if layer["type"] == "LUKS" and layer.get("data"):
                return DrbdVolumeData(layer["data"])
        return None

    @property
    def reports(self):
        return [ApiCallResponse(x) for x in self._rest_data.get("reports", [])]

    @property
    def data_v0(self):
        d = {
            "stor_pool_name": self.storage_pool_name,
            "vlm_nr": self.number,
            "vlm_uuid": self.uuid,
            "device_path": self.device_path
        }

        drbd_data = self.drbd_data
        if drbd_data is not None:
            d['vlm_minor_nr'] = drbd_data.drbd_volume_definition.minor
            if drbd_data.backing_device:
                d['backing_disk'] = drbd_data.backing_device
            d['meta_disk'] = drbd_data.meta_disk
            if drbd_data.allocated_size:
                d['allocated'] = self.allocated_size

        return d


class VolumeResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(VolumeResponse, self).__init__(rest_data)

    @property
    def volumes(self):
        """
        Resource volumes.
        :return: Resource volumes
        :rtype: list[Volume]
        """
        return list([Volume(x) for x in self._rest_data])


class Resource(RESTMessageResponse):
    def __init__(self, rest_data):
        super(Resource, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data["name"]

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def node_name(self):
        return self._rest_data["node_name"]

    @property
    def volumes(self):
        """
        Resource volumes.
        :return: Resource volumes
        :rtype: list[Volume]
        """
        return list([Volume(x) for x in self._rest_data.get("volumes", [])])

    @property
    def flags(self):
        """
        Resource flags as string list.

        :return: Resource definition flags as string list
        :rtype: list[str]
        """
        return self._rest_data.get("flags", [])

    @property
    def properties(self):
        """
        Resource properties.

        :return: Property map
        :rtype: dict[str, str]
        """
        return self._rest_data.get("props", {})

    @property
    def layer_data(self):
        """
        Return resource layer object
        :return:
        :rtype: ResourceLayerData
        """
        if "layer_object" in self._rest_data:
            return ResourceLayerData(self._rest_data["layer_object"])
        return None

    @property
    def create_datetime(self):
        """

        :return: Creation datetime of this resource
        :rtype: Optional[datetime]
        """
        if "create_timestamp" in self._rest_data:
            return datetime.fromtimestamp(self._rest_data["create_timestamp"] / 1000)
        return None

    @property
    def data_v0(self):
        return {
            "name": self.name,
            "uuid": self.uuid,
            "node_name": self.node_name,
            "rsc_flags": self.flags,
            "props": [{"key": x, "value": v} for x, v in self.properties.items()],
            "vlms": [x.data_v0 for x in self.volumes]
        }


class ResourceResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ResourceResponse, self).__init__(rest_data)

    @property
    def resources(self):
        """
        Return resource list from controller.
        :return: List of resources
        :rtype: list[Resource]
        """
        return list(Resource(x) for x in self._rest_data)

    @property
    def nodes(self):
        """
        List of node names, the resource is deployed
        :return:
        :rtype: list[str]
        """
        return [x["node_name"] for x in self._rest_data]

    @property
    def resource_states(self):
        """

        :return:
        :rtype: list[ResourceState]
        """
        return [ResourceState({
            "rsc_name": x["name"],
            "node_name": x["node_name"],
            "in_use": x.get("state", {}).get("in_use"),
            "vlm_states": [{
                "vlm_nr": y["volume_number"],
                "disk_state": y.get("state", {}).get("disk_state")
            } for y in x.get("volumes", [])]
        }) for x in self._rest_data]

    @property
    def data_v0(self):
        """
        Returns compatibility output for the first machine readable format.

        :return: Dictionary with old resource definition format
        """
        return {
            "resource_states": [x.data_v0 for x in self.resource_states],
            "resources": [x.data_v0 for x in self.resources]
        }


class ResourceConnection(RESTMessageResponse):
    def __init__(self, data):
        super(ResourceConnection, self).__init__(data)

    @property
    def node_a(self):
        return self._rest_data["node_a"]

    @property
    def node_b(self):
        return self._rest_data["node_b"]

    @property
    def flags(self):
        return self._rest_data.get("flags", [])

    @property
    def properties(self):
        return self._rest_data.get("props", {})

    @property
    def port(self):
        return self._rest_data.get("port")

    @property
    def data_v0(self):
        d = {
            "node_name_1": self.node_a,
            "node_name_2": self.node_b
        }
        if self.flags:
            d["flags"] = self.flags
        if self.properties:
            d["props"] = [{"key": x, "value": v} for x, v in self.properties.items()]
        if self.port is not None:
            d["port"] = self.port
        return d


class ResourceConnectionsResponse(RESTMessageResponse):
    def __init__(self, data):
        super(ResourceConnectionsResponse, self).__init__(data)

    @property
    def resource_connections(self):
        return [ResourceConnection(x) for x in self._rest_data]

    @property
    def data_v0(self):
        d = {}
        if self.resource_connections:
            d["rsc_connections"] = [x.data_v0 for x in self.resource_connections]
        return d


class SnapshotVolumeDefinition(RESTMessageResponse):
    def __init__(self, data):
        super(SnapshotVolumeDefinition, self).__init__(data)

    @property
    def number(self):
        return self._rest_data["volume_number"]

    @property
    def size(self):
        return self._rest_data["size_kib"]

    @property
    def data_v0(self):
        return {
            "vlm_nr": self.number,
            "vlm_size": self.size
        }


class Snapshot(RESTMessageResponse):
    def __init__(self, data):
        super(Snapshot, self).__init__(data)

    @property
    def name(self):
        return self._rest_data.get("snapshot_name")

    @property
    def node_name(self):
        return self._rest_data.get("node_name")

    @property
    def flags(self):
        """
        Resource flags as string list.

        :return: Resource definition flags as string list
        :rtype: list[str]
        """
        return self._rest_data.get("flags", [])

    @property
    def create_datetime(self):
        """

        :return: Creation datetime of this resource
        :rtype: Optional[datetime]
        """
        if "create_timestamp" in self._rest_data:
            return datetime.fromtimestamp(self._rest_data["create_timestamp"] / 1000)
        return None

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def snapshot_volumes(self):
        return [SnapshotVolume(x) for x in self._rest_data.get("snapshot_volumes", [])]


class SnapshotVolume(RESTMessageResponse):
    def __init__(self, data):
        super(SnapshotVolume, self).__init__(data)

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def vlm_nr(self):
        return self._rest_data.get("vlm_nr")

    @property
    def state(self):
        return self._rest_data.get("state")

    @property
    def properties(self):
        return self._rest_data.get("props", {})


class SnapshotDefinition(RESTMessageResponse):
    def __init__(self, data):
        super(SnapshotDefinition, self).__init__(data)

    @property
    def name(self):
        return self._rest_data["name"]

    @property
    def uuid(self):
        return self._rest_data.get("uuid")

    @property
    def snapshot_name(self):
        return self.name

    @property
    def resource_name(self):
        return self._rest_data["resource_name"]

    @property
    def rsc_name(self):
        return self.resource_name

    @property
    def nodes(self):
        """
        Node name list this snapshot is deployed.
        :return:
        :rtype: list[str]
        """
        return self._rest_data.get("nodes", [])

    @property
    def flags(self):
        return self._rest_data.get("flags", [])

    @property
    def snapshot_volume_definitions(self):
        return [SnapshotVolumeDefinition(x) for x in self._rest_data.get("volume_definitions", [])]

    @property
    def snapshots(self):
        return [Snapshot(x) for x in self._rest_data.get("snapshots", [])]

    @property
    def data_v0(self):
        return {
            "rsc_name": self.resource_name,
            "snapshot_name": self.snapshot_name,
            "uuid": self.uuid,
            "snapshot_dfn_flags": self.flags,
            "snapshots": [{"node_name": n} for n in self.nodes],
            "snapshot_vlm_dfns": [x.data_v0 for x in self.snapshot_volume_definitions]
        }


class SnapshotResponse(RESTMessageResponse):
    def __init__(self, data):
        super(SnapshotResponse, self).__init__(data)

    @property
    def snapshots(self):
        """
        Returns snapshot list
        :return:
        :rtype: list[SnapshotDefinition]
        """
        return [SnapshotDefinition(x) for x in self._rest_data]

    @property
    def data_v0(self):
        d = {}
        if self.snapshots:
            d["snapshot_dfns"] = [x.data_v0 for x in self.snapshots]
        return d


class Shipping(RESTMessageResponse):
    def __init__(self, data):
        super(Shipping, self).__init__(data)

    @property
    def snapshot_dfn(self):
        """
        Return the SnapshotDefinition object of the shipping
        :return: SnapshotDefinition object of ths shipping
        :rtype: SnapshotDefinition
        """
        return SnapshotDefinition(self._rest_data["snapshot"])

    @property
    def from_node_name(self):
        """
        Source node of the shipping
        :return: source node name
        :rtype: str
        """
        return self._rest_data["from_node_name"]

    @property
    def to_node_name(self):
        """
        Target node of the shipping
        :return: target node name
        :rtype: str
        """
        return self._rest_data["to_node_name"]

    @property
    def status(self):
        """
        Status of the shipping
        :return: status of the shipping
        :rtype: apiconsts.SnapshotShipStatus
        """
        return apiconsts.SnapshotShipStatus(self._rest_data["status"])


class SnapshotShippingResponse(RESTMessageResponse):
    def __init__(self, data):
        super(SnapshotShippingResponse, self).__init__(data)

    @property
    def shippings(self):
        """
        Returns snapshot shipping list
        :return:
        :rtype: list[Shipping]
        """
        return [Shipping(x) for x in self._rest_data]


class ControllerProperties(RESTMessageResponse):
    def __init__(self, data):
        super(ControllerProperties, self).__init__(data)

    @property
    def properties(self):
        return self._rest_data


class StoragePoolDefinition(RESTMessageResponse):
    def __init__(self, data):
        super(StoragePoolDefinition, self).__init__(data)

    @property
    def name(self):
        return self._rest_data["storage_pool_name"]

    @property
    def properties(self):
        return self._rest_data.get("props", {})

    @property
    def data_v0(self):
        d = {
            "stor_pool_name": self.name
        }

        if self.properties:
            d["props"] = [{"key": x, "value": v} for x, v in self.properties.items()]

        return d


class StoragePoolDefinitionResponse(RESTMessageResponse):
    def __init__(self, data):
        super(StoragePoolDefinitionResponse, self).__init__(data)

    @property
    def storage_pool_definitions(self):
        return [StoragePoolDefinition(x) for x in self._rest_data]

    @property
    def data_v0(self):
        d = {}
        if self.storage_pool_definitions:
            d["stor_pool_dfns"] = [x.data_v0 for x in self.storage_pool_definitions]
        return d


class Candidate(RESTMessageResponse):
    def __init__(self, data):
        super(Candidate, self).__init__(data)

    @property
    def max_volume_size(self):
        return self._rest_data["max_volume_size_kib"]

    @property
    def storage_pool(self):
        return self._rest_data["storage_pool"]

    @property
    def node_names(self):
        return self._rest_data.get("node_names", [])

    @property
    def all_thin(self):
        return self._rest_data["all_thin"]


class MaxVolumeSizeResponse(RESTMessageResponse):
    def __init__(self, data):
        super(MaxVolumeSizeResponse, self).__init__(data)

    @property
    def candidates(self):
        """

        :return:
        :rtype: list[Candidates]
        """
        return [Candidate(x) for x in self._rest_data.get("candidates", [])]

    @property
    def default_max_oversubscription_ratio(self):
        return self._rest_data["default_max_oversubscription_ratio"]


class ControllerVersion(RESTMessageResponse):
    def __init__(self, data):
        super(ControllerVersion, self).__init__(data)

    @property
    def version(self):
        return self._rest_data["version"]

    @property
    def git_hash(self):
        return self._rest_data.get("git_hash")

    @property
    def build_time(self):
        return self._rest_data["build_time"]

    @property
    def rest_api_version(self):
        return self._rest_data.get("rest_api_version", "1.0.0")


class NodeStorageEntry(RESTMessageResponse):
    def __init__(self, data):
        super(NodeStorageEntry, self).__init__(data)

    @property
    def device(self):
        return self._rest_data["device"]

    @property
    def model(self):
        return self._rest_data.get("model")

    @property
    def serial(self):
        return self._rest_data.get("serial")

    @property
    def wwn(self):
        return self._rest_data.get("wwn")


class PhysicalDevice(RESTMessageResponse):
    def __init__(self, data):
        super(PhysicalDevice, self).__init__(data)

    @property
    def size(self):
        return self._rest_data["size"]

    @property
    def rotational(self):
        return self._rest_data["rotational"]

    @property
    def nodes(self):
        """
        Returns a node map
        :return:
        :rtype: Dict[str, List[NodeStorageEntry]]
        """
        return {key: [NodeStorageEntry(x) for x in value] for key, value in self._rest_data.get("nodes", {}).items()}


class PhysicalStorageList(RESTMessageResponse):
    def __init__(self, data):
        super(PhysicalStorageList, self).__init__(data)

    @property
    def physical_devices(self):
        return [PhysicalDevice(x) for x in self._rest_data]


class SpaceReport(RESTMessageResponse):
    def __init__(self, data):
        super(SpaceReport, self).__init__(data)

    @property
    def report(self):
        return self._rest_data["reportText"]


class ExosDefaults(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosDefaults, self).__init__(rest_data)

    @property
    def username(self):
        return self._rest_data.get("username")

    @property
    def username_env(self):
        return self._rest_data.get("username_env")

    @property
    def password(self):
        return self._rest_data.get("password")

    @property
    def password_env(self):
        return self._rest_data.get("password_env")


class ExosEnclosure(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosEnclosure, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data["name"]

    @property
    def ctrl_a_ip(self):
        return self._rest_data.get("ctrl_a_ip")

    @property
    def ctrl_b_ip(self):
        return self._rest_data.get("ctrl_b_ip")

    @property
    def health(self):
        return self._rest_data["health"]

    @property
    def health_reason(self):
        return self._rest_data.get("health_reason")


class ExosListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosListResponse, self).__init__(rest_data)

    @property
    def exos_enclosures(self):
        """
        Returns a list with all EXOS enclosures.

        :return: The enclosure list.
        :rtype: list[ExosEnclosure]
        """
        return [ExosEnclosure(x) for x in self._rest_data]

    def exos_enclosure(self, encl_name):
        """
        Returns the specified enclosure from the list of enclosures.

        :param str encl_name: EXOS enclosure name
        :return: ExosEnclosure object of the enclosure, or None
        :rtype: ExosEnclosure
        """
        for encl in self.exos_enclosures:
            if encl.name == encl_name:
                return encl
        return None


class ExosEnclosureEvent(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosEnclosureEvent, self).__init__(rest_data)

    @property
    def severity(self):
        return self._rest_data["severity"]

    @property
    def event_id(self):
        return self._rest_data["event_id"]

    @property
    def controller(self):
        return self._rest_data["controller"]

    @property
    def time_stamp(self):
        return self._rest_data["time_stamp"]

    @property
    def time_stamp_numeric(self):
        return self._rest_data["time_stamp_numeric"]

    @property
    def message(self):
        return self._rest_data["message"]

    @property
    def additional_information(self):
        return self._rest_data["additional_information"]

    @property
    def recommended_action(self):
        return self._rest_data["recommended_action"]


class ExosEnclosureEventListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosEnclosureEventListResponse, self).__init__(rest_data)

    @property
    def exos_events(self):
        """
        Returns a list with the most current EXOS events.

        :return: The event list.
        :rtype: list[ExosEnclosureEvent]
        """
        return [ExosEnclosureEvent(x) for x in self._rest_data]


class ExosExecResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosExecResponse, self).__init__(rest_data)


class ExosMapResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosMapResponse, self).__init__(rest_data)

    @property
    def node_name(self):
        return self._rest_data['node_name']

    @property
    def enclosure_name(self):
        return self._rest_data['enclosure_name']

    @property
    def connections(self):
        return self._rest_data['connections']


class ExosMapListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExosMapListResponse, self).__init__(rest_data)

    @property
    def exos_connections(self):
        """
        Returns a list with currently active Linstor node <-> EXOS
        controller connections.

        :return: The map list.
        :rtype: list[ExosMapResponse]
        """
        return [ExosMapResponse(x) for x in self._rest_data]


class CloneStarted(RESTMessageResponse):
    def __init__(self, rest_data):
        super(CloneStarted, self).__init__(rest_data)

    @property
    def location(self):
        """
        API path for further status updates.
        :return: e.g. "/v1/resource-definitions/testrsc/clone/clonersc"
        :rtype: str
        """
        return self._rest_data["location"]

    @property
    def source_name(self):
        """
        Name of the source resource.

        :return:
        :rtype: str
        """
        return self._rest_data["source_name"]

    @property
    def clone_name(self):
        """
        Name of the cloned resource.

        :return:
        :rtype: str
        """
        return self._rest_data["clone_name"]

    @property
    def messages(self):
        """

        :return:
        :rtype: list[ApiCallResponse]
        """
        return [ApiCallResponse(x) for x in self._rest_data.get("messages", [])]


class CloneStatus(RESTMessageResponse):
    def __init__(self, rest_data):
        super(CloneStatus, self).__init__(rest_data)

    @property
    def status(self):
        """
        Current status of the clone operation.

        :return:
        :rtype: apiconsts.CloneStatus
        """
        return apiconsts.CloneStatus(self._rest_data["status"])


class SyncStatus(RESTMessageResponse):
    def __init__(self, rest_data):
        super(SyncStatus, self).__init__(rest_data)

    @property
    def synced_on_all(self):
        """
        Current sync status of the resource definition.

        :return: True if resource is ready to be e.g. resized
        :rtype: bool
        """
        return self._rest_data["synced_on_all"]


class S3Remote(RESTMessageResponse):
    def __init__(self, rest_data):
        super(S3Remote, self).__init__(rest_data)

    @property
    def remote_name(self):
        return self._rest_data["remote_name"]

    @property
    def endpoint(self):
        return self._rest_data["endpoint"]

    @property
    def bucket(self):
        return self._rest_data["bucket"]

    @property
    def region(self):
        return self._rest_data["region"]

    @property
    def access_key(self):
        return self._rest_data.get("access_key")

    @property
    def secret_key(self):
        return self._rest_data.get("secret_key")


class LinstorRemote(RESTMessageResponse):
    def __init__(self, rest_data):
        super(LinstorRemote, self).__init__(rest_data)

    @property
    def remote_name(self):
        return self._rest_data["remote_name"]

    @property
    def url(self):
        return self._rest_data["url"]

    @property
    def passphrase(self):
        return self._rest_data.get("passphrase")


class EbsRemote(RESTMessageResponse):
    def __init__(self, rest_data):
        super(EbsRemote, self).__init__(rest_data)

    @property
    def remote_name(self):
        return self._rest_data["remote_name"]

    @property
    def endpoint(self):
        return self._rest_data["endpoint"]

    @property
    def region(self):
        return self._rest_data.get("region")

    @property
    def availability_zone(self):
        return self._rest_data.get("availability_zone")


class RemoteListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(RemoteListResponse, self).__init__(rest_data)

    @property
    def s3_remotes(self):
        """

        :return:
        :rtype: list[S3Remote]
        """
        return [S3Remote(x) for x in self._rest_data.get("s3_remotes", [])]

    @property
    def linstor_remotes(self):
        """

        :return:
        :rtype: list[LinstorRemote]
        """
        return [LinstorRemote(x) for x in self._rest_data.get("linstor_remotes", [])]

    @property
    def ebs_remotes(self):
        """

        :return:
        :rtype: list[EbsRemote]
        """
        return [EbsRemote(x) for x in self._rest_data.get("ebs_remotes", [])]


class Backup(RESTMessageResponse):
    def __init__(self, rest_data):
        super(Backup, self).__init__(rest_data)

    @property
    def id(self):
        return self._rest_data.get("id")

    @property
    def start_time(self):
        return self._rest_data.get("start_time")

    @property
    def start_timestamp(self):
        return self._rest_data.get("start_timestamp")

    @property
    def finished_time(self):
        return self._rest_data.get("finished_time")

    @property
    def finished_timestamp(self):
        return self._rest_data.get("finished_timestamp")

    @property
    def origin_rsc_name(self):
        return self._rest_data.get("origin_rsc")

    @property
    def origin_snap_name(self):
        return self._rest_data.get("origin_snap")

    @property
    def based_on(self):
        return self._rest_data.get("based_on_id")

    @property
    def success(self):
        return self._rest_data.get("success")

    @property
    def shipping(self):
        return self._rest_data.get("shipping")

    @property
    def restorable(self):
        return self._rest_data.get("restorable")


class BackupOther(RESTMessageResponse):
    def __init__(self, rest_data):
        super(BackupOther, self).__init__(rest_data)

    @property
    def files(self):
        """
        Backup file list
        :return:
        :rtype: list[str]
        """
        return self._rest_data.get("files", [])


class BackupListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(BackupListResponse, self).__init__(rest_data)

    @property
    def linstor(self):
        """
        :return:
        :rtype: list[Backup]
        """
        return [Backup(v) for k, v in self._rest_data.get("linstor", {}).items()]

    @property
    def other(self):
        """
        :return:
        :rtype: list[BackupOther]
        """
        return BackupOther(self._rest_data.get("other", {}))


class BackupInfoResponse(RESTMessageResponse):

    def __init__(self, rest_data):
        super(BackupInfoResponse, self).__init__(rest_data)

    @property
    def rsc(self):
        return self._rest_data.get("rsc")

    @property
    def full(self):
        return self._rest_data.get("full")

    @property
    def snap(self):
        return self._rest_data.get("snap")

    @property
    def latest(self):
        return self._rest_data.get("latest")

    @property
    def count(self):
        return self._rest_data.get("count")

    @property
    def dl_size(self):
        return self._rest_data.get("dl_size_kib")

    @property
    def alloc_size(self):
        return self._rest_data.get("alloc_size_kib")

    @property
    def storpools(self):
        """
        :return:
        :rtype: list[BackupInfoStorPool]
        """
        return [BackupInfoStorPool(v) for v in self._rest_data.get("storpools", [])]


class BackupInfoStorPool(RESTMessageResponse):

    def __init__(self, rest_data):
        super(BackupInfoStorPool, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data.get("name")

    @property
    def provider_kind(self):
        return self._rest_data.get("provider_kind")

    @property
    def target_name(self):
        return self._rest_data.get("target_name")

    @property
    def remaining_space(self):
        return self._rest_data.get("remaining_space_kib")

    @property
    def volumes(self):
        """
        :return:
        :rtype: list[BackupInfoStorPoolVolume]
        """
        return [BackupInfoStorPoolVolume(v) for v in self._rest_data.get("vlms", [])]


class BackupInfoStorPoolVolume(RESTMessageResponse):

    def __init__(self, rest_data):
        super(BackupInfoStorPoolVolume, self).__init__(rest_data)

    @property
    def name(self):
        return self._rest_data.get("name")

    @property
    def layer_type(self):
        return self._rest_data.get("layer_type")

    @property
    def dl_size(self):
        return self._rest_data.get("dl_size_kib")

    @property
    def alloc_size(self):
        return self._rest_data.get("alloc_size_kib")

    @property
    def usable_size(self):
        return self._rest_data.get("usable_size_kib")


class ExternalFile(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ExternalFile, self).__init__(rest_data)

    @property
    def path(self):
        return self._rest_data.get("path")

    @property
    def content(self):
        return self._rest_data.get("content")


class FileResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(FileResponse, self).__init__(rest_data)

    @property
    def files(self):
        if isinstance(self._rest_data, list):
            return [ExternalFile(x) for x in self._rest_data]
        return ExternalFile(self._rest_data)


class Schedule(RESTMessageResponse):
    def __int__(self, rest_data):
        super(Schedule, self).__init__(rest_data)

    @property
    def schedule_name(self):
        return self._rest_data["schedule_name"]

    @property
    def full_cron(self):
        return self._rest_data["full_cron"]

    @property
    def inc_cron(self):
        return self._rest_data.get("inc_cron")

    @property
    def keep_local(self):
        return self._rest_data.get("keep_local")

    @property
    def keep_remote(self):
        return self._rest_data.get("keep_remote")

    @property
    def on_failure(self):
        return self._rest_data.get("on_failure")

    @property
    def max_retries(self):
        return self._rest_data.get("max_retries")


class ScheduleListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ScheduleListResponse, self).__init__(rest_data)

    @property
    def schedules(self):
        return [Schedule(x) for x in self._rest_data.get("data", [])]


class ScheduleResource(RESTMessageResponse):
    def __int__(self, rest_data):
        super(ScheduleResource, self).__init__(rest_data)

    @property
    def resource_name(self):
        return self._rest_data["rsc_name"]

    @property
    def remote_name(self):
        return self._rest_data.get("remote_name")

    @property
    def schedule_name(self):
        return self._rest_data.get("schedule_name")

    @property
    def reason(self):
        return self._rest_data.get("reason")

    @property
    def last_snap_time(self):
        """

        :return:
        :rtype: datetime
        """
        ts = self._rest_data.get("last_snap_time", -1)
        return None if ts < 0 else self._unix_epoch_field_value(ts)

    @property
    def last_snap_inc(self):
        """

        :return:
        :rtype: bool
        """
        return self._rest_data.get("last_snap_inc", False)

    @property
    def next_exec_time(self):
        """

        :return:
        :rtype: datetime
        """
        ts = self._rest_data.get("next_exec_time", -1)
        return None if ts < 0 else self._unix_epoch_field_value(ts)

    @property
    def next_exec_inc(self):
        """

        :return:
        :rtype: bool
        """
        return self._rest_data.get("next_exec_inc", False)

    @property
    def next_planned_full(self):
        """

        :return:
        :rtype: datetime
        """
        ts = self._rest_data.get("next_planned_full", -1)
        return None if ts < 0 else self._unix_epoch_field_value(ts)

    @property
    def next_planned_inc(self):
        """

        :return:
        :rtype: datetime
        """
        ts = self._rest_data.get("next_planned_inc", -1)
        return None if ts < 0 else self._unix_epoch_field_value(ts)


class ScheduleResourceListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ScheduleResourceListResponse, self).__init__(rest_data)

    @property
    def schedule_resources(self):
        return [ScheduleResource(x) for x in self._rest_data.get("data", [])]


class ScheduleResourceDetails(RESTMessageResponse):
    def __int__(self, rest_data):
        super(ScheduleResourceDetails, self).__init__(rest_data)

    @property
    def remote_name(self):
        return self._rest_data["remote_name"]

    @property
    def schedule_name(self):
        return self._rest_data["schedule_name"]

    @property
    def controller(self):
        return self._rest_data.get("ctrl")

    @property
    def resource_definition(self):
        return self._rest_data.get("rsc_dfn")

    @property
    def resource_group(self):
        return self._rest_data.get("rsc_grp")


class ScheduleResourceDetailsListResponse(RESTMessageResponse):
    def __init__(self, rest_data):
        super(ScheduleResourceDetailsListResponse, self).__init__(rest_data)

    @property
    def schedule_resources(self):
        return [ScheduleResourceDetails(x) for x in self._rest_data.get("data", [])]



#!/usr/bin/env python2

import sys
import linstor.linstorapi
from linstor.resource import _Client

if sys.version_info > (3, 0):
    unicode = str


class KV(dict):
    """
    KV class represents a Key-Value Store backed by LINSTOR.

    This object provides mainly the same interface as a Python3 dict and acts as a Key-Value store with
    namespaces. A namespace acts like a prefix in a UNIX file system and gets added to the every key that is
    set. The default namespace is '/'.

    Limitations: Keys and values have to be of type str. KV.keys(), KV.items() (as well as KV.values()) are
    implemented as simple generators. So they provide a "view", but not a "set-like" behavior. If you need
    that, convert the generator to a set (i.e. ks = KV.keys(); set(ks) & {'foo'}).

    The __repr__() method is not overridden intentionally and provides the representation of the whole dict,
    basically ignoring the namespace. This is helpful for debugging.

    Example:
        Add a value to a namespace /foo/bar, list it, change the namespace to "/" and list the item::

        $ kv = linstor.KV('myKV', namespace='/foo/bar/')
        $ kv['key'] = 'val'
        $ list(kv.items()) -> [('key', 'val')]
        $ kv.namespace = '/'
        $ list(kv.items()) -> [('/foo/bar/key', 'val')]
        $ kv['foo/baz/key'] = 'valbaz'
        $ kv.namespace = '/foo/bar'
        $ list(kv.items()) -> [('key', 'val')] # keys in /foo/baz not visible

    :param str name: The name of the KV-store. It acts as a unique handle.
    :param str namespace: A UNIX-like file system path.
    :param bool cached: A KV can be backed by a python dict as a caching layer. As of now has to be True.
     This influences whether reads are forwarded to LINSTOR or not. Writes are obviously always forwarded.
    :param str uri: A list of controller addresses.
    :param bool rw_to_linstor: If set to False, entries are not written to LINSTOR. It can be used to use
     the KV as a name spaced dict() only, or for debugging.
    :param linstor.Linstor existing_client: Instead of creating  a new client based on the controller addresses,
     use this pre-configured client object.
    """
    def __init__(self, name, namespace='/', cached=True, uri='linstor://localhost', rw_to_linstor=True,
                 existing_client=None):
        assert cached, 'Currently we only allow "cached" KeyValueStores'
        assert cached or rw_to_linstor, 'KV has to be "cached" and/or "rw_to_linstor"'
        super(KV, self).__init__()
        self._name = name
        self._cached = cached
        self.client = _Client(uri)
        self._existing_client = existing_client
        self._rw_to_linstor = rw_to_linstor
        self._import()
        self._set_ns(namespace)

    def _get_connection(self):
        if self._existing_client:
            return self._existing_client
        return linstor.MultiLinstor(self.client.uri_list, self.client.timeout, self.client.keep_alive)

    # keys are expected to be expanded
    def _set_linstor_kv(self, k, v):
        if not self._rw_to_linstor:
            return
        with self._get_connection() as lin:
            rs = lin.keyvaluestore_modify(self._name, property_dict={k: v}, delete_props=None)
            if not rs[0].is_success():
                raise linstor.LinstorError('Could not set kv({}:{}): {}'.format(k, v, rs[0]))

    # keys are expected to be expanded
    # we could check for an iterable, but a str is also iterable, be explicit, we only have one list-caller
    # still, we want that special case for a list, it is a lot more efficient (see clear()).
    def _del_linstor_kv(self, k, is_list_like=False):
        if not self._rw_to_linstor:
            return

        if is_list_like:
            delete_props = list(k)  # allows e.g., tuples
        else:
            delete_props = [k]
        with self._get_connection() as lin:
            rs = lin.keyvaluestore_modify(self._name, property_dict=None, delete_props=delete_props)
            if not rs[0].is_success():
                raise linstor.LinstorError('Could not delete kv({}): {}'.format(k, rs[0]))

    @classmethod
    def _valid_string(cls, s):
        if isinstance(s, str):
            return True
        else:
            return isinstance(s, unicode)

    @classmethod
    def _normalize_ns(cls, ns):
        ns = ns.strip()
        while True:
            if ns.startswith('/'):
                ns = ns[1:].lstrip()
            elif ns.endswith('/'):
                ns = ns[:-1].rstrip()
            else:
                break

        return ns

    def _import(self):
        d = {}
        if self._rw_to_linstor:
            with self._get_connection() as lin:
                d = {'/' + k: v for k, v in lin.keyvaluestore_list(self._name).properties.items()}

        super(KV, self).clear()
        super(KV, self).update(d)

    # works for None, '', False, '/', and proper namespaces
    def _set_ns(self, ns):
        if ns:
            ns = KV._normalize_ns(ns)
        ns = '/{}/'.format(ns) if ns else '/'
        self._ns = ns

    def _key_ns_add(self, k):
        return self._ns + self._normalize_ns(k)

    def _key_ns_del(self, k):
        return k[len(self._ns):]

    @property
    def namespace(self):
        """
        Returns the current name space (e.g., /foo/).

        :return: Current name space.
        :rtype: str
        """
        return self._ns

    @namespace.setter
    def namespace(self, ns):
        self._set_ns(ns)

    def __delitem__(self, k):
        if not KV._valid_string(k):
            raise KeyError('key {} has to be a str/unicode, but is {}'.format(k, type(k)))
        k = self._key_ns_add(k)
        self._del_linstor_kv(k)
        super(KV, self).__delitem__(k)

    def __setitem__(self, k, v):
        if not KV._valid_string(k):
            raise KeyError('key {} has to be a str/unicode, but is {}'.format(k, type(k)))
        if not KV._valid_string(v):
            raise ValueError('value {} has to be a str/unicode, but is {}'.format(v, type(v)))

        k = self._key_ns_add(k)
        self._set_linstor_kv(k, v)
        super(KV, self).__setitem__(k, v)

    def __getitem__(self, k):
        return super(KV, self).__getitem__(self._key_ns_add(k))

    def __contains__(self, k):
        return super(KV, self).__contains__(self._key_ns_add(k))

    def __iter__(self):
        for k in super(KV, self).__iter__():
            if k.startswith(self.namespace):
                yield self._key_ns_del(k)

    def clear(self):
        to_delete = [k for k in super(KV, self).keys() if k.startswith(self.namespace)]
        if len(to_delete) > 0:
            self._del_linstor_kv(to_delete, is_list_like=True)
            for k in to_delete:
                super(KV, self).__delitem__(k)

    def get(self, k, d=None):
        return super(KV, self).get(self._key_ns_add(k), d)

    def items(self):
        for k, v in super(KV, self).items():
            if k.startswith(self.namespace):
                yield self._key_ns_del(k), v

    def keys(self):
        return self.__iter__()

    def pop(self, *args):
        k = args[0]
        if self.__contains__(k):
            item = self.__getitem__(k)
            self.__delitem__(k)
            return item

        if len(args) == 2:  # default given
            return args[1]

        raise KeyError

    def popitem(self):
        kv = list(self.items())
        if len(kv) == 0:
            raise KeyError

        k, v = kv[-1]
        self.__delitem__(k)
        return k, v

    def setdefault(self, k, d=None):
        if self.__contains__(k):
            return self.__getitem__(k)
        self.__setitem__(k, d)
        return d

    def update(self, *args, **kwargs):
        raise NotImplementedError

    def values(self):
        for _, v in self.items():
            yield v



#!/usr/bin/env python3
"""
    linstor - management of distributed DRBD9 resources
    Copyright (C) 2013 - 2018  LINBIT HA-Solutions GmbH
    Author: Robert Altnoeder, Philipp Reisner

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import re
import subprocess
import codecs
from setuptools import setup, Command
from setuptools.command.build_py import build_py


def get_version():
    """
    Function to parse the version string from the linstor/__init__.py.
    This was done to not rely on a working build, e.g. sometimes files were
    not yet generated and this made problems with importing.

    :return: semantic version string
    :rtype: str
    """
    with open('linstor/version.py') as linstor_init:
        for line in linstor_init:
            if line.startswith('VERSION'):
                m = re.search(r'"(.*)"', line)
                if m:
                    return m.group(1)
                else:
                    raise RuntimeError("Unable to parse version: " + line)
    raise RuntimeError("Unable to find version string.")


# used to overwrite version tag by internal build tools
# keep it, even if you don't understand it.
def get_setup_version():
    return get_version()


class CheckUpToDate(Command):
    description = "Check if version strings are up to date"
    user_options = []

    def initialize_options(self):
        self.cwd = None

    def finalize_options(self):
        self.cwd = os.getcwd()

    def run(self):
        version = get_version()
        try:
            with codecs.open("debian/changelog", encoding='utf8', errors='ignore') as f:
                firstline = f.readline()
                if version not in firstline:
                    # returning false is not promoted
                    sys.exit(1)
        except IOError:
            # probably a release tarball without the debian directory but with Makefile
            return True


class BuildPyCommand(build_py):
    """
    Run make gensrc before doing build.
    """
    def run(self):
        subprocess.check_call(["make", "gensrc"])
        build_py.run(self)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="python-linstor",
    version=get_version(),
    description="Linstor python api",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://www.linbit.com',
    project_urls={
        "Source Code": "https://github.com/LINBIT/linstor-api-py",
        "Documentation": "https://linbit.github.io/linstor-api-py",
    },
    author="Robert Altnoeder <robert.altnoeder@linbit.com>, Roland Kammerer <roland.kammerer@linbit.com>"
           + ", Rene Peinthor <rene.peinthor@linbit.com>, Moritz Wanzenboeck <moritz.wanzenboeck@linbit.com>",
    author_email="rene.peinthor@linbit.com",
    maintainer="LINBIT HA-Solutions GmbH",
    maintainer_email="drbd-user@lists.linbit.com",
    license="LGPLv3",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    extras_require={
        ":python_version<'3'": ['enum34'],
    },
    packages=[
        'linstor'
    ],
    # package_data={},
    cmdclass={
        "versionup2date": CheckUpToDate,
        "build_py": BuildPyCommand
    },
    test_suite="linstor_tests"
)



#!/usr/bin/env python3
from __future__ import print_function

import argparse
import json
import os
import sys
import subprocess
import datetime

script_path = os.path.dirname(os.path.abspath(__file__))
CONSTS_JSON_PATH = os.path.join(script_path, 'consts.json')

basename = os.path.basename(sys.argv[0])
COPY_YEAR = datetime.datetime.fromtimestamp(os.path.getmtime(CONSTS_JSON_PATH)).year
hdr = 'This file was autogenerated by %s' % basename
license_text = """
LINSTOR - management of distributed storage/DRBD9 resources
Copyright (C) 2017 - %s  LINBIT HA-Solutions GmbH
All Rights Reserved.
Author: %s

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
  License for the specific language governing permissions and limitations
  under the License.""" % (
    COPY_YEAR, ', '.join(['Robert Altnoeder', 'Roland Kammerer', 'Gabor Hernadi', 'Rene Peinthor']))


def get_native_java_type(type_str):
    """
    Return the native java type.

    :param str type_str:
    :return:
    """
    if type_str == 'BOR':
        return 'long'
    elif type_str == 'BAND':
        return 'long'
    if type_str == 'string':
        return 'String'
    elif type_str == 'bool':
        return 'boolean'
    elif type_str == 'int':
        return 'int'
    elif type_str == 'long':
        return 'long'

    raise NotImplementedError()


def java(consts, outdir):
    indent = ' ' * 4

    license_hdr = ''
    for line in license_text.split('\n'):
        license_hdr += (' * ' + line).rstrip() + '\n'

    with open(os.path.join(outdir, "ApiConsts.java"), "w+") as outfile:
        outfile.write('/*\n * %s\n%s */\n' % (hdr, license_hdr) + '\n')

        outfile.write('package com.linbit.linstor.api;\n\n')
        outfile.write('@SuppressWarnings({"checkstyle:magicnumber", "checkstyle:constantname"})\n')
        outfile.write('public class ApiConsts\n{\n')

        nl, w = '', 0
        for e in consts:
            w += 1
            if w > 1:
                nl = '\n'

            if 'blockcomment' in e:
                c = e['blockcomment'].replace('\n', '\n' + indent + ' * ')
                outfile.write('%s%s/*\n %s* %s\n %s*/\n' % (nl, indent, indent, c, indent))
                continue

            _type = e['type']
            if _type == 'enum':
                enum_type = e['enumtype']
                native_type = get_native_java_type(enum_type)
                outfile.write('%spublic enum %s\n%s{\n' % (indent, e['name'], indent))
                enum_strs = []
                for enum_entry in e['values']:
                    enum_value = '"%s"' % enum_entry['value'] if enum_type == 'string' else enum_entry['value']
                    enum_strs.append('%s%s(%s)' % (indent * 2, enum_entry['name'], str(enum_value)))
                outfile.write(",\n".join(enum_strs) + ';\n\n')

                outfile.write("%sprivate final %s enumValue;\n\n" % (indent * 2, native_type))

                outfile.write("%s%s(final %s val)\n" % (indent * 2, e['name'], native_type))
                outfile.write("%s{\n" % (indent * 2))
                outfile.write("%senumValue = val;\n" % (indent * 3))
                outfile.write("%s}\n\n" % (indent * 2))

                outfile.write("%spublic %s getValue()\n" % (indent * 2, native_type))
                outfile.write("%s{\n" % (indent * 2))
                outfile.write("%sreturn enumValue;\n" % (indent * 3))
                outfile.write("%s}\n\n" % (indent * 2))

                outfile.write('%s}\n' % indent)
            else:
                value = e['value']
                native_type = get_native_java_type(_type)

                if native_type is None:
                    raise RuntimeError("Type '{t}' not handled.".format(t=_type))

                if _type == 'BOR':
                    value = ' | '.join([str(x) for x in value])
                elif _type == 'BAND':
                    value = ' & '.join([str(x) for x in value])
                elif _type == 'string':
                    value = '"%s"' % value

                if 'comment' in e:
                    outfile.write("%s// %s\n" % (indent, e['comment']))
                c = "%spublic static final %s %s = %s;" % (indent, native_type, e['name'], value)
                outfile.write(c + '\n')

        outfile.write('\n    private ApiConsts()\n    {\n    }\n')
        outfile.write('}\n')


def strip_l(value):
    if str(value).endswith('L') and (str(value).startswith('0x') or value[:-1].isnumeric()):
        value = value[:-1]
    return value


def get_native_python_type(type_str):
    """

    :param str type_str:
    :return:
    :rtype: str
    """
    if type_str == 'BOR':
        return 'long'
    elif type_str == 'BAND':
        return 'long'
    elif type_str == 'string':
        return 'str'
    elif type_str == 'bool':
        return 'bool'
    elif type_str == 'int':
        return 'int'
    elif type_str == 'long':
        return 'long'

    raise NotImplementedError()


def python(consts, outdir):
    with open(os.path.join(outdir, "sharedconsts.py"), "w+") as outfile:
        outfile.write('# %s\n' % hdr)
        for line in license_text.split('\n'):
            outfile.write(('# ' + line).rstrip() + '\n')
        outfile.write("""import sys
from enum import Enum

if sys.version_info > (3,):
    long = int

""")

        store_mask = True
        masks = []
        indent = ' ' * 4
        nl, w = '', 0
        for e in consts:
            w += 1
            nl = '\n' if w > 1 else ''

            if 'blockcomment' in e:
                c = e['blockcomment'].replace('\n', '\n# ')
                if 'Special answer message content types' in c:
                    store_mask = False
                outfile.write('\n%s# ## %s ###\n' % (nl, c))
                continue

            _type = e['type']
            if _type == 'enum':
                outfile.write('\n\nclass %s(Enum):\n' % (e['name']))
                for enum_entry in e['values']:
                    fmt_str = '%s%s = "%s"' if e['enumtype'] == 'string' else '%s%s = %s'
                    outfile.write(fmt_str % (indent, enum_entry['name'], enum_entry['value']) + '\n')
            else:
                value = e['value']

                if _type in ['long', 'BOR', 'BAND']:
                    if isinstance(value, list):
                        value = [strip_l(x) for x in value]
                    else:
                        value = strip_l(value)

                native_type = get_native_python_type(_type)

                if _type == 'BOR':
                    value = ' | '.join([str(x) for x in value])
                elif _type == 'BAND':
                    value = ' & '.join([str(x) for x in value])
                elif _type == 'string':
                    value = "'%s'" % value

                if store_mask:
                    masks.append(e['name'])

                assert (native_type is not None)
                if 'comment' in e:
                    outfile.write("# %s\n" % (e['comment']))
                c = "%s = %s(%s)" % (e['name'], native_type, value)
                outfile.write(c + '\n')

        outfile.write('\n\nif __name__ == "__main__":\n')
        outfile.write(' ' * 4 + 'MAP_MASK = {\n')
        outfile.write(',\n'.join([' ' * 8 + "'" + mask + "': " + mask for mask in masks]) + '\n')
        outfile.write(' ' * 4 + '}\n')
        outfile.write("""    TYPE_MASKS = [
        0xC000000000000000,  # TYPE
        0x0000000003000000,  # OPERATION
        0x00000000003C0000,  # OBJECT
        0xC00000000000FFFF   # ACTION
    ]

    for num_str in sys.argv[1:]:
        num = long(num_str)
        mask = []
        for type_mask in TYPE_MASKS:
            for key, mask_value in MAP_MASK.items():
                if num & type_mask == mask_value:
                    mask.append(key)
        print(num_str + " = " + " | ". join(mask))
""")


def to_camel_case(snake_str):
    components = snake_str.split('_')  # split by underscore and .title() items
    return ''.join(x.title() for x in components)


def golang(consts, outdir):
    apiconsts_path = os.path.join(outdir, "apiconsts.go")
    files_written = [apiconsts_path]
    with open(apiconsts_path, "w+") as outfile:
        outfile.write('// %s\n\n' % hdr)
        outfile.write('// '.join([line.strip() + '\n' for line in license_text.split('\n')]) + '\n')

        outfile.write('package linstor\n')

        store_mask = True
        masks = []
        nl, w = '', 0
        translated = {}

        for e in consts:
            w += 1
            nl = '\n' if w > 1 else ''

            if 'blockcomment' in e:
                c = e['blockcomment'].replace('\n', '\n// ')
                if 'Special answer message content types' in c:
                    store_mask = False
                outfile.write('%s// ## %s ###\n' % (nl, c))
                continue

            _type = e['type']

            if _type == 'enum':
                package_name = e['name'].lower()
                enum_type_name = e['name']
                outfile.write('// enum generated in package -> "golinstor/{p}"\n'.format(p=package_name))
                subpkg = os.path.join(outdir, package_name)
                os.makedirs(subpkg, exist_ok=True)
                subpkg_file_path = os.path.join(subpkg, package_name + '.go')
                files_written.append(subpkg_file_path)
                with open(subpkg_file_path, "w+") as subfile:
                    subfile.write("package {p}\n\n".format(p=package_name))

                    subfile.write("type {p} {t}\n\n".format(p=enum_type_name, t=e['enumtype']))
                    subfile.write("const (\n")
                    for enum_entry in e['values']:
                        enum_name = to_camel_case(enum_entry['name'])
                        enum_value = '"%s"' % enum_entry['value'] if e['enumtype'] == 'string' else enum_entry['value']
                        outfile.write('// {p}.{en} = {v}\n'.format(
                            p=package_name, en=enum_name, v=enum_value))
                        subfile.write("    {en} {t} = {v}\n".format(
                            en=enum_name, t=enum_type_name, v=enum_value))
                    subfile.write(")\n")
            else:
                value = e['value']
                if _type in ['long', 'BOR', 'BAND']:
                    if isinstance(value, list):
                        value = [strip_l(x) for x in value]
                    else:
                        value = strip_l(value)

                if _type == 'BOR':
                    value = [translated.get(v, v) for v in value]
                    value = ' | '.join([str(x) for x in value])
                    value = '(%s)' % value
                elif _type == 'BAND':
                    value = [translated.get(v, v) for v in value]
                    value = ' & '.join([str(x) for x in value])
                    value = '(%s)' % value
                elif _type == 'string':
                    value = '"%s"' % value

                if store_mask:
                    masks.append(e['name'])

                if 'comment' in e:
                    outfile.write("// %s\n" % (e['comment']))
                var = snake_to_camel(e['name'])
                translated[e['name']] = var
                c = "const %s = %s" % (var, value)
                outfile.write(c + '\n')

    # gofmt files
    for filepath in files_written:
        print('gofmt file ' + filepath)
        subprocess.check_call(['gofmt', '-w', filepath])


def snake_to_camel(name):
    return "".join(w.lower().title() for w in name.split("_"))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("language", choices=['java', 'python', 'golang'])
    parser.add_argument("outdir", default=".")

    args = parser.parse_args()

    with open(CONSTS_JSON_PATH) as consts_file:
        try:
            consts = json.load(consts_file)
        except Exception as e:
            print('The input file (%s) is not valid, better luck next time...\n' % CONSTS_JSON_PATH, file=sys.stderr)
            print('Error: %s...\n' % e, file=sys.stderr)
            sys.exit(1)

    if args.language == 'java':
        java(consts, args.outdir)
    elif args.language == 'python':
        python(consts, args.outdir)
    elif args.language == 'golang':
        golang(consts, args.outdir)


if __name__ == "__main__":
    main()



import os

from linstor.linstorapi import MultiLinstor

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser as ConfigParser


class Config(object):
    @staticmethod
    def read_config(config_file):
        cp = ConfigParser()
        cp.read(config_file)
        config = {}
        for section in cp.sections():
            config[section] = cp.items(section)
        return config

    @staticmethod
    def get_section(section, config_file_name=None):
        home_dir = os.path.expanduser("~")
        config_file = "linstor-client.conf"
        user_conf = os.path.join(home_dir, ".config", "linstor", config_file)
        sys_conf = os.path.join('/etc', 'linstor', config_file)

        config = None
        if config_file_name and os.path.exists(config_file_name):
            config = Config.read_config(config_file_name)
        elif os.path.exists(user_conf):
            config = Config.read_config(user_conf)
        elif os.path.exists(sys_conf):
            config = Config.read_config(sys_conf)

        entries = config.get(section, []) if config else []
        return {k: v for k, v in entries}

    @staticmethod
    def get_controllers(section='global', config_file_name=None, fallback='linstor://localhost'):
        """
        :param str section: Section to parse, defaults to 'global'
        :param str config_file_name: Config file to parse, if not set, the default files in /etc and $HOME are
         parsed
        :param str fallback: Fallback controller if none was found in the environment or the config. Defaults
         to 'linstor://localhost'.
        :return: List of linstor uris. This list is intended to be used via the MultiLinstor class.
        :rtype: list[str]
        """
        controllers = os.environ.get('LS_CONTROLLERS', None)
        if not controllers:
            cfg = Config.get_section(section, config_file_name)
            controllers = cfg.get('controllers', fallback)

        return MultiLinstor.controller_uri_list(controllers)
